"""
An example of the django-synchronise project.
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 't9l&8)%_k0(6lhk)e5d2r$k&@lc#mtcs(8(665dvpux-%*k9g+'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'synchronise',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'example.urls'

WSGI_APPLICATION = 'example.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

STATIC_URL = '/static/'

TEMPLATE_DIRS = (
    os.path.join(BASE_DIR, 'templates'),
)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '[%(levelname)-5s] [%(asctime)s] %(filename)s:'
                      '%(lineno)03d | %(message)s',
            'datefmt': '%Y.%m.%d %H:%M:%S'
        },
        'simple': {
            'format': '[%(levelname)-5s] [%(filename)s:%(lineno)03d] |'
                      ' %(message)s'
        },
    },
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'propagate': True,
            'level': 'INFO',
        },
        'synchronise': {
            'handlers': ['console'],
            'propagate': True,
            'level': 'DEBUG',
        }
    }
}



from django.http import HttpResponse
from django.views.generic import View

from synchronise import synchronise

import json


class SynchroniseView(View):

    def post(self, request, *args, **kwargs):
        """
        Handle the synchronise request.  Gets the JSON payload and check
        for the user name and project name.  Then the synchroniser is called.
        """
        try:
            payload_string = request.POST['payload']
            payload_json = json.loads(payload_string)
            user = request.GET.get('user')
            project = request.GET.get('project')
            return synchronise.synchronise(payload_json, user, project)
        except ValueError:
            return HttpResponse('Post contains an invalid JSON payload.\n',
                                status=400, reason='Invalid JSON payload.')
        except KeyError:
            return HttpResponse('A JSON payload needs to be provided.\n',
                                status=400, reason='No JSON payload.')



__version_info__ = {
    'major': 0,
    'minor': 2,
    'micro': 0,
    'releaselevel': 'final',
    'serial': 42
}


def get_version(short=False):
    assert __version_info__['releaselevel'] in ('alpha', 'beta', 'final')
    version = ["{}.{}.{}".format(__version_info__['major'],
                                 __version_info__['minor'],
                                 __version_info__['micro'])]
    if __version_info__['releaselevel'] != 'final' and not short:
        version.append('{}{}'.format(__version_info__['releaselevel'][0],
                                     __version_info__['serial']))
    return ''.join(version)

__version__ = get_version()



from django.conf.urls import patterns, url
from django.views.decorators.csrf import csrf_exempt
from synchronise.views import SynchroniseView

urlpatterns = patterns(
    '',
    url(r'^', csrf_exempt(SynchroniseView.as_view()), name='synchronise'),
)



from django.conf.urls import patterns, include, url

from django.contrib import admin
admin.autodiscover()

urlpatterns = patterns(
    '',
    url(r'^synchronise/', include('synchronise.urls')),
    url(r'^admin/', include(admin.site.urls))
)



from django.test import Client
from urllib.parse import quote_plus

import unittest

import synchronise


class VersionTest(unittest.TestCase):

    def test_version_plain(self):
        self.assertEqual(synchronise.get_version(), '0.2.0')

    def test_version_short(self):
        self.assertEqual(synchronise.get_version(), '0.2.0')


class BitBucketToGitHubTest(unittest.TestCase):

    def setUp(self):
        self.client = Client()

    def test_empty_post(self):
        response = self.client.post('/synchronise/', {})
        self.assertEqual(response.status_code, 400)

    def test_empty_post_with_user(self):
        response = self.client.post('/synchronise/?user=jw', {})
        self.assertEqual(response.status_code, 400)

    def test_invalid_json_posts(self):
        content_type = 'application/x-www-form-urlencoded'
        # prepare the first request (no JSON payload)
        invalid_json = '{ "Hello": "There" }'
        invalid_json_url = quote_plus(invalid_json)
        payload = b'payload=' + bytes(invalid_json_url, 'utf-8')
        # send it off
        response = self.client.post('/synchronise/?user=jw',
                                    payload, content_type=content_type)
        # check
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase, 'No proper JSON in payload.')
        # prepare the second request (invalid JSON payload)
        invalid_json = '{ "Invalid": "No empty quote }'
        invalid_json_url = quote_plus(invalid_json)
        payload = b'payload=' + bytes(invalid_json_url, 'utf-8')
        # send it off
        response = self.client.post('/synchronise/?user=jw',
                                    payload, content_type=content_type)
        # check
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.reason_phrase, 'Invalid JSON payload.')

    def test_valid_post1(self):
        """
        Test the push of this repository to the same repository on GitHub.
        """
        valid_post = {
            'payload': '{                                        '
                       '   "canon_url": "https://bitbucket.org", '
                       '   "repository": {                       '
                       '       "scm": "hg",                      '
                       '       "owner": "elevenbits",            '
                       '       "name": "django-synchroniser"     '
                       '   },                                    '
                       '   "user": "elevenbits"                  '
                       '}                                        '
        }
        response = self.client.post('/synchronise/',
                                    valid_post)
        self.assertEqual(response.status_code, 200)

    def test_valid_post2(self):
        """
        Test the push of this repository to the same repository on GitHub.
        """
        valid_post = {
            'payload': '{                                        '
                       '   "canon_url": "https://bitbucket.org", '
                       '   "repository": {                       '
                       '       "scm": "hg",                      '
                       '       "owner": "elevenbits",            '
                       '       "name": "django-synchroniser"     '
                       '   },                                    '
                       '   "user": "elevenbits"                  '
                       '}                                        '
        }
        user_project = 'user=jw&project=django-synchronise'
        response = self.client.post('/synchronise/?' + user_project,
                                    valid_post)
        self.assertEqual(response.status_code, 200)



from django.http import HttpResponse

import configparser
from os.path import join
from tempfile import TemporaryDirectory

import hgapi

import logging
logger = logging.getLogger(__name__)


def print_payload(payload, length):
    """
    Convert the payload to a string, and replace each whitespace section
    with a single space. Return it with a specified length.
    :param payload: A dict.
    :param length: The required length.
    :return: The processed payload of the specified length.
    """
    clean = ' '.join(str(payload).split())
    if len(clean) <= length:
        return clean
    else:
        return '"' + clean[0:length] + '..."'


def is_bitbucket_hg_payload(payload):
    """
    Check to see of the payload is a BitBucket POST from a Mercurial based
    project
    :param payload: The payload as a dist
    :return: True when it is a BitBucket POST from a Mercurial payload; False
    otherwise.
    """

    logger.debug('Checking {} for bitbucket and hg...'.
                 format(print_payload(payload, 20)))
    result = False
    try:
        if 'bitbucket' in payload['canon_url']:
            repo = payload['repository']
            if repo['scm'] == 'hg':
                logger.debug('This is a Bitbucket Mercurial repository POST.')
                result = True
    except KeyError:
        pass  # no canon_url or scm in payload
    return result


def bitbucket_hg_payload(payload, project=None, user=None):
    """
    Prepare a BitBucket Mercurial based project for synchronisation via
    HgGit
    :param payload: The payload as a dict
    :param project: The project string
    :param user: The userstring
    :return:
    """
    repo = payload['repository']
    # use default user and project when not given
    if user is None:
        user = repo['owner']
    if project is None:
        project = repo['name']
    # hg_path should be ssh://hg@bitbucket.org/<name>/<project>,
    hg_path = 'ssh://hg@bitbucket.org/{}/{}'.format(repo['owner'],
                                                    repo['name'])
    # git_path should be git+ssh://git@github.com/<user>/<project>.git
    git_path = 'git+ssh://git@github.com/{}/{}.git'.format(user, project)
    try:
        hg_to_git(hg_path, git_path)
    except hgapi.HgException as he:
        HttpResponse("Could not convert: {}.".format(he),
                     status=500, reason='HgApi problem: {}'.format(he))
    message = "Converted {} successfully".format(project)
    return HttpResponse(message, status=200, reason=message)


def synchronise(payload, user=None, project=None):
    """
    Synchronised the given payload.
    :param payload: The payload as a dict
    :param user: The user
    :param project: The project
    :return: A HttpResponse, either 200 when successful, 400 or 500 otherwise.
    """
    if is_bitbucket_hg_payload(payload):
        return bitbucket_hg_payload(payload, project, user)
    return HttpResponse('No proper JSON in payload. Discarding it.',
                        status=400, reason='No proper JSON in payload.')


def add_hggit_extension_and_git_path(hg_repo_path, git_path):
    """
    Add hggit extension to the hgrc and add git_path to its paths section.
    This is required to inform git of the hg changesets on the hg repository.
    :param hg_repo_path: The path of the hgrc
    :param git_path: The path of the github repo
    :return:
    """
    hgrc = join(hg_repo_path, '.hg', 'hgrc')
    config = configparser.ConfigParser()
    config.read(hgrc)
    # add hgit extension
    if 'extensions' in config.sections():
        if 'hggit' not in config['extensions']:
            config['extensions']['hggit']
    else:
        config['extensions'] = {}
        config['extensions']['hggit'] = ''
    # add git_path to the paths section
    if 'paths' in config.sections():
        if 'github' not in config['paths']:
            config['paths']['github'] = git_path
    else:
        config['paths'] = {}
        config['paths'][git_path] = ''
    # write the hgrc
    with open(hgrc, 'w') as config_file:
        config.write(config_file)


def hg_to_git(hg_path, git_path):
    """
    Convert the Bitbucket repo to a Github repo. First clone the hg_path,
    then add the hggit extension to this clone, and lastly push the clone
    to github.
    :param hg_path: The Bitbucket Mercurial based repository.
    :param git_path: The Github (obviously Git based) repository.
    :return:
    """
    logger.debug('Converting {} to {}.'.format(hg_path, git_path))
    with TemporaryDirectory(prefix='hg-') as hg_repo_path:
        logger.debug('Cloning {} to {}...'.format(hg_path, hg_repo_path))
        hg_repo = hgapi.hg_clone(hg_path, hg_repo_path)
        add_hggit_extension_and_git_path(hg_repo_path, git_path)
        logger.debug('Bookmarking the clone...')
        hg_repo.hg_bookmarks(action=hgapi.Repo.BOOKMARK_CREATE, name='master')
        logger.debug('Pushing it to github...')
        hg_repo.hg_push('github')


def git_to_hg(git_path, hg_path):
    logger.debug('Converting {} to {}.'.format(git_path, hg_path))
    raise NotImplemented


def git_to_git(git_path, other_git_path):
    logger.debug('Converting {} to {}.'.format(git_path, other_git_path))
    raise NotImplemented



from setuptools import setup, find_packages

kwargs = {
    'packages': find_packages(exclude=['tests', '*.tests',
                                       '*.tests.*', 'tests.*']),
    'include_package_data': True,
    'install_requires': [
        'django>=1.6,<1.7',
    ],
    'name': 'django-synchroniser',
    'version': __import__('synchronise').get_version(),
    'author': 'Jan Willems',
    'author_email': 'jw@elevenbits.com',
    'description': 'django-synchroniser automagically synchronises Bitbucket '
                   'Mecurial base projects to GitHub (Git based) projects',
    'license': 'GPLv3',
    'keywords': 'synchronise bitbucket github',
    'url': 'https://github.com/elevenbits/django-synchroniser',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
}

setup(**kwargs)
